import datetime
import typing

import pydantic

from ._base_api import BaseAPIClient


class Account(pydantic.BaseModel):
    id: str
    name: str


class WhoAmIAPIKey(pydantic.BaseModel):
    id: str
    account: Account


class WhoAmICaller(pydantic.BaseModel):
    api_key: WhoAmIAPIKey


class WhoAmIResponse(pydantic.BaseModel):
    caller: WhoAmICaller


class Evaluator(pydantic.BaseModel):
    id: str
    name: str
    evaluator_family: str | None
    aliases: list[str] | None


class ListEvaluatorsResponse(pydantic.BaseModel):
    evaluators: list[Evaluator]


class Project(pydantic.BaseModel):
    id: str
    name: str


class CreateProjectRequest(pydantic.BaseModel):
    name: str


class Experiment(pydantic.BaseModel):
    project_id: str
    id: str
    name: str


class CreateExperimentRequest(pydantic.BaseModel):
    project_id: str
    name: str


class CreateExperimentResponse(pydantic.BaseModel):
    experiment: Experiment


class EvaluateEvaluator(pydantic.BaseModel):
    evaluator: str
    profile_name: str | None = None
    explain_strategy: str = "always"


class EvaluateRequest(pydantic.BaseModel):
    evaluators: list[EvaluateEvaluator]
    evaluated_model_system_prompt: str | None = None
    evaluated_model_retrieved_context: list[str] | None = None
    evaluated_model_input: str | None = None
    evaluated_model_output: str | None = None
    evaluated_model_gold_answer: str | None = None
    experiment_id: str
    capture: str = "all"
    dataset_id: str | None = None
    dataset_sample_id: int | None = None
    tags: dict[str, str] | None = None


class EvaluationResultAdditionalInfo(pydantic.BaseModel):
    positions: list | None
    extra: dict | None
    confidence_interval: dict | None


class EvaluationResult(pydantic.BaseModel):
    id: str
    project_id: str
    experiment_id: str
    created_at: pydantic.AwareDatetime
    evaluator_id: str
    evaluated_model_system_prompt: str | None
    evaluated_model_retrieved_context: list[str] | None
    evaluated_model_input: str | None
    evaluated_model_output: str | None
    evaluated_model_gold_answer: str | None
    pass_: bool | None = pydantic.Field(alias="pass")
    score_raw: float | None
    additional_info: EvaluationResultAdditionalInfo
    explanation: str | None
    evaluation_duration: datetime.timedelta | None
    explanation_duration: datetime.timedelta | None
    evaluator_family: str
    evaluator_profile_public_id: str
    dataset_id: str | None
    dataset_sample_id: int | None
    tags: dict[str, str] | None


class EvaluateResult(pydantic.BaseModel):
    evaluator_id: str
    profile_name: str
    status: str
    error_message: str | None
    evaluation_result: EvaluationResult | None


class EvaluateResponse(pydantic.BaseModel):
    results: list[EvaluateResult]


class ExportEvaluationResult(pydantic.BaseModel):
    experiment_id: str
    evaluator_id: str
    profile_name: str | None = None
    evaluated_model_system_prompt: str | None = None
    evaluated_model_retrieved_context: list[str] | None = None
    evaluated_model_input: str | None = None
    evaluated_model_output: str | None = None
    evaluated_model_gold_answer: str | None = None
    pass_: bool | None = pydantic.Field(alias="pass_", serialization_alias="pass")
    score_raw: float | None
    evaluation_duration: datetime.timedelta | None = None
    evaluated_model_name: str | None = None
    evaluated_model_provider: str | None = None
    evaluated_model_params: dict[str, str | int | float] | None = None
    evaluated_model_selected_model: str | None = None
    dataset_id: str | None = None
    dataset_sample_id: int | None = None
    tags: dict[str, str] | None = None


class ExportEvaluationRequest(pydantic.BaseModel):
    evaluation_results: list[ExportEvaluationResult]


class ExportEvaluationResultPartial(pydantic.BaseModel):
    id: str
    app: str | None
    created_at: pydantic.AwareDatetime
    evaluator_id: str


class ExportEvaluationResponse(pydantic.BaseModel):
    evaluation_results: list[ExportEvaluationResultPartial]


class ListProfilesRequest(pydantic.BaseModel):
    public_id: str | None = None
    evaluator_family: str | None = None
    evaluator_id: str | None = None
    name: str | None = None
    revision: str | None = None
    get_last_revision: bool = False
    is_patronus_managed: bool | None = None
    limit: int = 1000
    offset: int = 0


class EvaluatorProfile(pydantic.BaseModel):
    public_id: str
    evaluator_family: str
    name: str
    revision: int
    config: dict[str, typing.Any] | None
    is_patronus_managed: bool
    created_at: datetime.datetime
    description: str | None


class CreateProfileRequest(pydantic.BaseModel):
    evaluator_family: str
    name: str
    config: dict[str, typing.Any]


class CreateProfileResponse(pydantic.BaseModel):
    evaluator_profile: EvaluatorProfile


class AddEvaluatorProfileRevisionRequest(pydantic.BaseModel):
    config: dict[str, typing.Any]


class AddEvaluatorProfileRevisionResponse(pydantic.BaseModel):
    evaluator_profile: EvaluatorProfile


class ListProfilesResponse(pydantic.BaseModel):
    evaluator_profiles: list[EvaluatorProfile]


class DatasetDatum(pydantic.BaseModel):
    dataset_id: str
    sid: int
    evaluated_model_system_prompt: str | None = None
    evaluated_model_retrieved_context: list[str] | None = None
    evaluated_model_input: str | None = None
    evaluated_model_output: str | None = None
    evaluated_model_gold_answer: str | None = None
    meta_evaluated_model_name: str | None = None
    meta_evaluated_model_provider: str | None = None
    meta_evaluated_model_selected_model: str | None = None
    meta_evaluated_model_params: dict[str, str | int | float] | None = None


class ListDatasetData(pydantic.BaseModel):
    data: list[DatasetDatum]


class API(BaseAPIClient):
    async def whoami(self) -> WhoAmIResponse:
        resp = await self.call("GET", "/v1/whoami", response_cls=WhoAmIResponse)
        resp.response.raise_for_status()
        return resp.data

    async def create_project(self, request: CreateProjectRequest) -> Project:
        resp = await self.call("POST", "/v1/projects", body=request, response_cls=Project)
        resp.response.raise_for_status()
        return resp.data

    async def create_experiment(self, request: CreateExperimentRequest) -> Experiment:
        resp = await self.call("POST", "/v1/experiments", body=request, response_cls=CreateExperimentResponse)
        resp.response.raise_for_status()
        return resp.data.experiment

    async def evaluate(self, request: EvaluateRequest) -> EvaluateResponse:
        resp = await self.call("POST", "/v1/evaluate", body=request, response_cls=EvaluateResponse)
        # TODO handle the error better
        for res in resp.data.results:
            if res.status != "success":
                raise RuntimeError(f"Unexpected response from /v1/evaluate: {res}")
        # TODO error handling
        resp.response.raise_for_status()
        return resp.data

    async def export_evaluations(self, request: ExportEvaluationRequest) -> ExportEvaluationResponse:
        resp = await self.call(
            "POST",
            "/v1/evaluation-results/batch",
            body=request,
            response_cls=ExportEvaluationResponse,
        )
        # TODO error handling
        resp.response.raise_for_status()
        return resp.data

    async def list_evaluators(self) -> list[Evaluator]:
        resp = await self.call("GET", "/v1/evaluators", response_cls=ListEvaluatorsResponse)
        # TODO error handling
        resp.response.raise_for_status()
        return resp.data.evaluators

    async def create_profile(self, request: CreateProfileRequest) -> CreateProfileResponse:
        resp = await self.call("POST", "/v1/evaluator-profiles", body=request, response_cls=CreateProfileResponse)
        # TODO error handling
        resp.response.raise_for_status()
        return resp.data

    async def add_evaluator_profile_revision(
        self, evaluator_profile_id, request: AddEvaluatorProfileRevisionRequest
    ) -> AddEvaluatorProfileRevisionResponse:
        resp = await self.call(
            "POST",
            f"/v1/evaluator-profiles/{evaluator_profile_id}/revision",
            body=request,
            response_cls=AddEvaluatorProfileRevisionResponse,
        )
        # TODO error handling
        resp.response.raise_for_status()
        return resp.data

    async def list_profiles(self, request: ListProfilesRequest) -> ListProfilesResponse:
        params = request.model_dump(exclude_none=True)
        resp = await self.call(
            "GET",
            "/v1/evaluator-profiles",
            params=params,
            response_cls=ListProfilesResponse,
        )
        # TODO error handling
        resp.response.raise_for_status()
        return resp.data

    async def list_dataset_data(self, dataset_id: str) -> ListDatasetData:
        resp = await self.call("GET", f"/v1/datasets/{dataset_id}/data", response_cls=ListDatasetData)
        # TODO error handling
        resp.response.raise_for_status()
        return resp.data



import inspect
import logging
import asyncio
import datetime
import itertools
import os
import re
import statistics
import time
import typing
import urllib.parse
from concurrent.futures import ThreadPoolExecutor

from ._config import config
from ._dataset import DatasetDatum, Dataset
from ._async_utils import run_until_complete
from . import _api as api
from ._client import Client
from ._evaluators import Evaluator, EvaluatorOutput
from ._tasks import Task, TaskResult, nop_task

log = logging.getLogger(__name__)


class DefaultReporter:
    # batch_size 10 is max batch size that Patronus AI API accepts in single call
    batch_size = 10

    def __init__(self, client: Client, experiment_id: str, flush_interval: int = 10):
        self._client = client
        self.experiment_id = experiment_id

        self._lock = asyncio.Lock()
        self.last_export_slice = slice(0, 0)
        self.last_flush_ts = time.time()
        self.flush_interval = flush_interval

        self.evaluators = set()
        self.remote_results: list[api.ExportEvaluationResult] = []
        self.outgoing_results: list[api.ExportEvaluationResult] = []

    async def add(
        self,
        evaluator_name: str,
        profile_name: str | None,
        input_datum: dict,
        task_result: TaskResult,
        evaluation_result: EvaluatorOutput,
        tags: dict[str, str],
        *,
        dataset_id: str,
        dataset_sample_id: int,
        already_captured: bool,
    ):
        entry = api.ExportEvaluationResult(
            experiment_id=self.experiment_id,
            evaluator_id=evaluator_name,
            profile_name=profile_name,
            evaluated_model_system_prompt=task_result.evaluated_model_system_prompt
            or input_datum.get("evaluated_model_system_prompt"),
            evaluated_model_retrieved_context=input_datum.get("evaluated_model_retrieved_context"),
            evaluated_model_input=input_datum.get("evaluated_model_input"),
            evaluated_model_output=task_result.evaluated_model_output,
            evaluated_model_gold_answer=input_datum.get("evaluated_model_gold_answer"),
            pass_=evaluation_result.result.pass_,
            score_raw=evaluation_result.result.score_raw,
            evaluation_duration=datetime.timedelta(seconds=evaluation_result.duration),
            evaluated_model_name=task_result.evaluated_model_name,
            evaluated_model_provider=task_result.evaluated_model_provider,
            evaluated_model_params=task_result.evaluated_model_params,
            evaluated_model_selected_model=task_result.evaluated_model_selected_model,
            dataset_id=dataset_id,
            dataset_sample_id=dataset_sample_id,
            tags=tags,
        )
        async with self._lock:
            self.evaluators.add((evaluator_name, profile_name))
            if already_captured:
                self.remote_results.append(entry)
            else:
                self.outgoing_results.append(entry)

        await self._conditional_flush()

    async def _conditional_flush(self):
        async with self._lock:
            buffered = len(self.outgoing_results) - self.last_export_slice.stop
            if buffered == 0:
                return
            if buffered >= self.batch_size or self.last_flush_ts + self.flush_interval < time.time():
                await self._flush()

    async def flush(self):
        async with self._lock:
            buffered = len(self.outgoing_results) - self.last_export_slice.stop
            if buffered == 0:
                return
            await self._flush()

    async def _flush(self):
        while self.last_export_slice.stop < len(self.outgoing_results):
            upper_idx = min(
                len(self.outgoing_results),
                self.last_export_slice.stop + self.batch_size,
            )
            self.last_export_slice = slice(self.last_export_slice.stop, upper_idx)
            results_to_export = self.outgoing_results[self.last_export_slice]
            await self._client.api.export_evaluations(
                api.ExportEvaluationRequest(
                    evaluation_results=results_to_export,
                )
            )
            log.debug(f"Flushed {len(results_to_export)} results")
        self.last_flush_ts = time.time()

    def summary(self):
        for evaluator_name, profile_name in self.evaluators:
            name = evaluator_name if not profile_name else f"{evaluator_name}:{profile_name}"
            results: typing.Iterator[api.ExportEvaluationResult] = itertools.chain(
                self.outgoing_results, self.remote_results
            )
            results = filter(
                lambda r: r.evaluator_id == evaluator_name and r.profile_name == profile_name,
                results,
            )
            scores_and_passes = list(map(lambda r: (r.score_raw, r.pass_), results))

            scores = [x[0] for x in scores_and_passes if x[0] is not None]
            passes = [int(x[1]) for x in scores_and_passes if x[1] is not None]

            print_summary(name, scores, passes, len(scores_and_passes), display_hist=True)


async def with_semaphore(sem, task):
    async with sem:
        return await task


class Experiment:
    project_name: str
    project_id: str | None
    experiment_id: str | None
    # TODO handle with API?
    experiment_name: str
    whoami: api.WhoAmIResponse | None

    dataset: Dataset | None
    task: Task
    evaluators: list[Evaluator] | None

    _client: Client | None

    _pool: ThreadPoolExecutor
    _sem: asyncio.Semaphore

    def __init__(
        self,
        client: Client | None,
        project_name: str | None,
        data: list[DatasetDatum] | typing.Callable[[...], Dataset] | typing.Awaitable[Dataset],
        task: Task,
        evaluators: list[Evaluator | typing.Awaitable[Evaluator]],
        tags: dict[str, str],
        max_concurrency: int,
        experiment_name: str = "",
        **kwargs,
    ):
        self._client = client
        self.project_id = None
        self.project_name = re.sub(r"[^a-zA-Z0-9\-_]", "-", project_name) or "default"
        self.experiment_id = None
        self.experiment_name = generate_experiment_name(experiment_name)

        self.dataset = None
        self.__data = data
        self.task = task
        self.evaluators = None
        self.__evaluators = evaluators

        self.tags = tags

        self._sem = asyncio.Semaphore(max_concurrency)
        self._pool = ThreadPoolExecutor()

        self.reporter = None

    async def prepare(self):
        print("Preparing dataset... ", end="")
        self.dataset = await self.fetch_dataset()
        print("DONE")

        print("Preparing evaluators... ", end="")
        self.evaluators = await self.prepare_evaluators()
        print("DONE")

        if not self._client:
            return

        self.whoami = await self._client.api.whoami()

        project = await self._client.api.create_project(api.CreateProjectRequest(name=self.project_name))
        self.project_id = project.id
        self.project_name = project.name

        ex = await self._client.api.create_experiment(
            api.CreateExperimentRequest(project_id=self.project_id, name=self.experiment_name)
        )
        self.experiment_id = ex.id

        # TODO associate reported that doesn't need client if client not available
        self.reporter = DefaultReporter(self._client, self.experiment_id, flush_interval=10)

    async def fetch_dataset(self) -> (str | None, list[DatasetDatum]):
        if isinstance(self.__data, (list, tuple)):
            dataset = self.__data
        elif inspect.iscoroutine(self.__data):
            dataset = await self.__data
        elif inspect.iscoroutinefunction(self.__data):
            dataset = await self.__data()
        elif callable(self.__data):
            return self.__data()
        else:
            raise ValueError("'data' passed to the experiment is an unexpected object")

        if isinstance(dataset, Dataset):
            return dataset
        return Dataset(dataset_id=None, data=dataset)

    async def prepare_evaluators(self) -> list[Evaluator]:
        evaluators = []
        for ev in self.__evaluators:
            if inspect.iscoroutine(ev):
                evaluators.append(await ev)
            elif inspect.iscoroutinefunction(ev):
                evaluators.append(await ev())
            else:
                evaluators.append(ev)
        return evaluators

    async def run(self):
        title = f"Running experiment: {self.project_name}/{self.experiment_name}"
        print("=" * len(title))
        print(title)
        print("=" * len(title))

        tasks = []
        for i, datum in enumerate(self.dataset.data, 1):
            sample_id = datum.get("sid", i)
            task = self.run_task_and_eval(datum, dataset_id=self.dataset.dataset_id, dataset_sample_id=sample_id)
            tasks.append(asyncio.create_task(with_semaphore(self._sem, task)))

        for t in tasks:
            await t

        await self.reporter.flush()

        self.reporter.summary()

        print()
        print(get_link(self.whoami.caller.api_key.account.id, self.experiment_id))

    async def run_task_and_eval(self, datum, dataset_id: str | None, dataset_sample_id: int):
        loop = asyncio.get_running_loop()

        em_system_prompt = datum.get("evaluated_model_system_prompt")
        em_retrieved_context = datum.get("evaluated_model_retrieved_context")
        em_input = datum.get("evaluated_model_input")
        em_output = datum.get("evaluated_model_output")
        em_gold_answer = datum.get("evaluated_model_gold_answer")

        task = await self.task.execute(
            loop,
            self._pool,
            evaluated_model_system_prompt=em_system_prompt,
            evaluated_model_retrieved_context=em_retrieved_context,
            evaluated_model_input=em_input,
            evaluated_model_output=em_output,
            evaluated_model_gold_answer=em_gold_answer,
            tags=self.tags,
        )

        outgoing_tags = self.tags
        if task.tags:
            outgoing_tags = {**self.tags, **task.tags}

        futures = [
            loop.create_task(
                evaluator.execute(
                    loop,
                    self._pool,
                    experiment_id=self.experiment_id,
                    evaluated_model_system_prompt=em_system_prompt,
                    evaluated_model_retrieved_context=em_retrieved_context,
                    evaluated_model_input=em_input,
                    evaluated_model_output=task.evaluated_model_output,
                    evaluated_model_gold_answer=em_gold_answer,
                    dataset_id=dataset_id,
                    dataset_sample_id=dataset_sample_id,
                    tags=outgoing_tags,
                )
            )
            for evaluator in self.evaluators
        ]

        for evaluator, f in zip(self.evaluators, futures):
            eval_result = await f

            await self.reporter.add(
                evaluator.name,
                evaluator.profile_name,
                datum,
                task,
                eval_result,
                outgoing_tags,
                dataset_id=dataset_id,
                dataset_sample_id=dataset_sample_id,
                already_captured=evaluator.remote_capture,
            )


def experiment(
    client: Client | None,
    project_name: str | None,
    *,
    data: list[DatasetDatum] | typing.Callable[[...], list[DatasetDatum]] | typing.Awaitable[list[DatasetDatum]],
    task: Task = nop_task,
    evaluators: list[Evaluator],
    tags: dict[str, str] | None = None,
    experiment_name: str = "",
    **kwargs,
):
    ex = Experiment(
        client=client,
        project_name=project_name,
        data=data,
        task=task,
        evaluators=evaluators,
        tags=tags or {},
        max_concurrency=10,
        experiment_name=experiment_name,
        **kwargs,
    )

    async def run():
        await ex.prepare()
        await ex.run()

    return run_until_complete(run())


def print_summary(name: str, scores: list[float], passes: list[int], count: int, display_hist: bool):
    title = f"Summary: {name}"

    print()
    print(title)
    print("-" * len(title))
    print(f"Count     : {count}")
    print(f"Pass rate : {round(statistics.mean(passes), 3)}")
    print(f"Mean      : {round(statistics.mean(scores), 3)}")
    print(f"Min       : {round(min(scores), 3)}")
    print(f"25%       : {round(percentile(scores, 25), 3)}")
    print(f"50%       : {round(percentile(scores, 50), 3)}")
    print(f"75%       : {round(percentile(scores, 75), 3)}")
    print(f"Max       : {round(max(scores), 3)}")

    if display_hist:
        print()
        print("Score distribution")
        print_histogram(scores)


def gen_name(name: str) -> str:
    name = re.sub(r"[^a-zA-Z0-9]", "", name).lower()
    ts = datetime.datetime.now().strftime("%y%m%d%H%M%S")
    name = name or "unknown"
    return f"ex-{name}-{ts}"


def percentile(data: list[float], p: int):
    data = sorted(data)
    index = (p / 100) * (len(data) - 1)
    if index.is_integer():
        return data[int(index)]
    else:
        lower_bound = int(index)
        upper_bound = lower_bound + 1
        weight = index - lower_bound
        return data[lower_bound] * (1 - weight) + data[upper_bound] * weight


def print_histogram(data, bin_count=5):
    # Calculate the range of the data
    min_val = min(data)
    max_val = max(data)

    if min_val == max_val:
        if min_val > 0.5:
            min_val = 0
        else:
            max_val = 1

    range_val = max_val - min_val

    # Calculate bin size
    bin_size = range_val / bin_count

    # Initialize bins
    bins = [0] * bin_count

    # Distribute data into bins
    for value in data:
        # Find the appropriate bin for the current value
        bin_index = int((value - min_val) / bin_size)
        # Edge case for the maximum value
        if bin_index == bin_count:
            bin_index -= 1
        bins[bin_index] += 1

    # Determine the width of the histogram
    max_bin_count = max(bins)
    scale_factor = 20 / max_bin_count  # Scale the histogram to a max width of 50 characters

    # Print the histogram
    print("Score Range".ljust(20), "Count".ljust(10), "Histogram")
    for i in range(bin_count):
        bin_start = min_val + i * bin_size
        bin_end = bin_start + bin_size
        bin_count = bins[i]
        bar = "#" * int(bin_count * scale_factor)
        print(f"{bin_start:.2f} - {bin_end:.2f}".ljust(20), f"{bin_count}".ljust(10), bar)


def get_link(account_id: str, experiment_id: str) -> str:
    params = {"account_id": account_id, "experiment_id": experiment_id}
    ui_url = config().ui_url.rstrip("/")
    return f"{ui_url}/monitoring?{urllib.parse.urlencode(params)}"


def generate_experiment_name(name: str) -> str:
    ts = int(time.time())
    if name:
        return f"{name}-{ts}"
    try:
        login = os.getlogin()
        return f"{login}-{ts}"
    except OSError:  # Possible in-cluster error: No such device or address
        return str(ts)



import dataclasses
import typing


class DatasetDatum(typing.TypedDict):
    sid: typing.NotRequired[int | None]
    evaluated_model_system_prompt: typing.NotRequired[str | None]
    evaluated_model_retrieved_context: typing.NotRequired[list[str] | None]
    evaluated_model_input: typing.NotRequired[str | None]
    evaluated_model_output: typing.NotRequired[str | None]
    evaluated_model_gold_answer: typing.NotRequired[str | None]


@dataclasses.dataclass
class Dataset:
    dataset_id: str | None
    data: list[DatasetDatum]



import logging
import typing

import httpx

from ._config import config
from ._dataset import Dataset, DatasetDatum
from ._evaluators import Evaluator
from ._evaluators_remote import RemoteEvaluator
from ._tasks import Task, nop_task
from . import _api as api

log = logging.getLogger(__name__)


class Client:
    def __init__(
        self,
        api_key: str | None = None,
        base_url: str = "",
        api_client: api.API | None = None,
        # TODO Allow passing more types for the timeout: float, Timeout, None, NotSet
        timeout: float = 300,
    ):
        api_key = api_key or config().api_key
        base_url = base_url or config().api_url

        if not api_key:
            raise ValueError("Provide 'api_key' argument or set PATRONUSAI_API_KEY environment variable.")

        if api_client is None:
            # TODO allow passing http client as an argument
            http_client = httpx.AsyncClient(timeout=timeout)

            # TODO use package version
            api_client = api.API(version="0.0.1", http=http_client)
        api_client.set_target(base_url, api_key)
        self.api = api_client

    def experiment(
        self,
        project_name: str,
        *,
        data: list[DatasetDatum] | typing.Callable[[...], list[DatasetDatum]] | typing.Awaitable[list[DatasetDatum]],
        task: Task = nop_task,
        evaluators: list[Evaluator],
        tags: dict[str, str] | None = None,
        experiment_name: str = "",
        **kwargs,
    ):
        from ._experiment import experiment as ex

        ex(
            self,
            project_name=project_name,
            data=data,
            task=task,
            evaluators=evaluators,
            tags=tags,
            experiment_name=experiment_name,
            **kwargs,
        )

    async def remote_evaluator(
        self,
        # ID or an alias of an evaluator.
        evaluator: str,
        # profile_name is not necessary for evaluators that not requires them, like "toxicity".
        profile_name: str | None = None,
        *,
        profile_config: dict[str, typing.Any] | None = None,
        allow_update: bool = False,
        explain_strategy: typing.Literal["never", "on-fail", "on-success", "always"] = "always",
    ) -> RemoteEvaluator:
        evaluators = await self.api.list_evaluators()

        ev: api.Evaluator | None = None
        for e in evaluators:
            if e.id == evaluator:
                ev = e
            for alias in e.aliases:
                if alias == evaluator:
                    ev = e

        if ev is None:
            raise ValueError(f"Evaluator {evaluator!r} not found")

        if profile_config:
            return await self._remote_evaluator_from_config(
                evaluator=evaluator,
                profile_name=profile_name,
                ev=ev,
                profile_config=profile_config,
                allow_update=allow_update,
                explain_strategy=explain_strategy,
            )
        else:
            return await self._remote_evaluator(
                evaluator=evaluator,
                profile_name=profile_name,
                ev=ev,
                explain_strategy=explain_strategy,
            )

    async def _remote_evaluator_from_config(
        self,
        *,
        evaluator: str,
        profile_name: str,
        ev: api.Evaluator,
        profile_config: dict[str, typing.Any],
        allow_update: bool,
        explain_strategy: typing.Literal["never", "on-fail", "on-success", "always"],
    ) -> RemoteEvaluator:
        if not profile_name:
            raise ValueError("profile_name is required when specifying profile_config")
        if profile_name.startswith("system:"):
            raise ValueError(f"Cannot use profile_config with system profiles. Provided profile was {profile_name!r}")

        profiles = (
            await self.api.list_profiles(
                api.ListProfilesRequest(
                    evaluator_family=ev.evaluator_family,
                    name=profile_name,
                    get_last_revision=True,
                )
            )
        ).evaluator_profiles

        if not profiles:
            log.info(
                f"No evaluator profile {profile_name!r} for evaluator {ev.evaluator_family!r} found. Creating one..."
            )
            profile = await self._create_profile_from_config(ev.evaluator_family, profile_name, profile_config)
            log.info(f"Evaluator profile {profile_name} created for evaluator family {ev.evaluator_family}.")
        elif len(profiles) > 1:
            raise Exception(
                f"Unexpected number of profiles retrieved for "
                f"evaluator {evaluator!r} and profile name {profile_name!r}"
            )
        else:
            profile = profiles[0]

        # Check if user provided profile config is subset of existing config
        # This checks only one level of the config, but we don't support profiles with nested
        # structure at this point so, it's alright.
        is_subset = {**profile.config, **profile_config} == profile.config

        if not is_subset and not allow_update:
            raise ValueError(
                "Provided 'profile_config' differs from existing profile. "
                "Please set 'allow_update=True' if you wish to update the profile. "
                "Updating profiles can be unsafe if they're used in production system or by other people."
            )

        if not is_subset:
            log.info("Existing profile config differs from the provided config. Adding revision to the profile...")
            profile = (
                await self.api.add_evaluator_profile_revision(
                    profile.public_id,
                    api.AddEvaluatorProfileRevisionRequest(
                        config={**profile.config, **profile_config},
                    ),
                )
            ).evaluator_profile
            log.info(f"Revision added to evaluator profile {profile_name}.")

        return RemoteEvaluator(
            evaluator=ev.id,
            profile_name=profile.name,
            explain_strategy=explain_strategy,
            api_=self.api,
        )

    async def _create_profile_from_config(
        self,
        evaluator_family: str,
        profile_name: str,
        profile_config: dict[str, typing.Any],
    ) -> api.EvaluatorProfile:
        resp = await self.api.create_profile(
            api.CreateProfileRequest(
                evaluator_family=evaluator_family,
                name=profile_name,
                config=profile_config,
            )
        )
        return resp.evaluator_profile

    async def _remote_evaluator(
        self,
        *,
        evaluator: str,
        profile_name: str | None,
        ev: api.Evaluator,
        explain_strategy: typing.Literal["never", "on-fail", "on-success", "always"],
    ) -> RemoteEvaluator:
        profiles = await self.api.list_profiles(
            api.ListProfilesRequest(
                evaluator_family=ev.evaluator_family,
                name=profile_name,
                get_last_revision=True,
            )
        )
        if len(profiles.evaluator_profiles) == 0:
            raise ValueError(f"Profile for evaluator {evaluator!r} given name {profile_name!r} not found")
        if len(profiles.evaluator_profiles) > 1:
            raise ValueError(f"More than 1 profile found for evaluator {evaluator!r}")

        profile = profiles.evaluator_profiles[0]

        return RemoteEvaluator(
            evaluator=ev.id,
            profile_name=profile.name,
            explain_strategy=explain_strategy,
            api_=self.api,
        )

    async def remote_dataset(self, dataset_id: str) -> Dataset:
        resp = await self.api.list_dataset_data(dataset_id)
        return Dataset(dataset_id=dataset_id, data=resp.model_dump()["data"])



import typing

from . import _evaluators as evaluators
from . import _api as api


class RemoteEvaluatorError(evaluators.EvaluatorError):
    def __init__(self, evaluator: str, profile_name: str, status: str, error_message: str):
        super().__init__(
            f"{evaluator!r} with profile {profile_name!r} "
            f"returned unexpected status {status!r} with error message {error_message!r}"
        )


class RemoteEvaluator(evaluators.Evaluator):
    remote_capture = True

    def __init__(
        self,
        evaluator: str,
        profile_name: str,
        explain_strategy: typing.Literal["never", "on-fail", "on-success", "always"],
        api_: api.API,
    ):
        self.name = evaluator
        self.evaluator = evaluator
        self.profile_name = profile_name
        self.explain_strategy = explain_strategy
        self.api = api_

        super().__init__(evaluators.EVALUATION_ARGS)

    async def evaluate(
        self,
        experiment_id: str,
        evaluated_model_system_prompt: str | None = None,
        evaluated_model_retrieved_context: list[str] | None = None,
        evaluated_model_input: str | None = None,
        evaluated_model_output: str | None = None,
        evaluated_model_gold_answer: str | None = None,
        dataset_id: str | None = None,
        dataset_sample_id: int | None = None,
        tags: dict[str, str] | None = None,
    ) -> evaluators.EvaluationResultT:
        # TODO error handling
        response = await self.api.evaluate(
            api.EvaluateRequest(
                evaluators=[
                    api.EvaluateEvaluator(
                        evaluator=self.evaluator,
                        profile_name=self.profile_name,
                        explain_strategy=self.explain_strategy,
                    )
                ],
                evaluated_model_system_prompt=evaluated_model_system_prompt,
                evaluated_model_retrieved_context=evaluated_model_retrieved_context,
                evaluated_model_input=evaluated_model_input,
                evaluated_model_output=evaluated_model_output,
                evaluated_model_gold_answer=evaluated_model_gold_answer,
                experiment_id=experiment_id,
                capture="all",
                dataset_id=dataset_id,
                dataset_sample_id=dataset_sample_id,
                tags=tags,
            )
        )
        data = response.results[0]
        if data.status != "success":
            raise RemoteEvaluatorError(self.evaluator, self.profile_name, data.status, data.error_message)

        return data.evaluation_result



import abc
import asyncio
import dataclasses
import inspect
import typing
from concurrent.futures import ThreadPoolExecutor

from ._async_utils import run_as_coro

TASK_ARGS = {
    "evaluated_model_system_prompt",
    "evaluated_model_retrieved_context",
    "evaluated_model_input",
    "evaluated_model_output",
    "evaluated_model_gold_answer",
    "tags",
}


class TaskResultT(typing.Protocol):
    evaluated_model_output: str
    evaluated_model_system_prompt: str | None
    evaluated_model_name: str | None
    evaluated_model_provider: str | None
    evaluated_model_params: str | None
    evaluated_model_selected_model: str | None
    tags: dict[str, str] | None


@dataclasses.dataclass
class TaskResult:
    evaluated_model_output: str
    evaluated_model_system_prompt: str | None = None
    evaluated_model_name: str | None = None
    evaluated_model_provider: str | None = None
    evaluated_model_params: dict[str, int | float | str] | None = None
    evaluated_model_selected_model: str | None = None
    tags: dict[str, str] | None = None


TaskFn = typing.Callable[..., TaskResultT | str]


class Task(abc.ABC):
    def __init__(self, accepted_args: set[str]):
        self.accepted_args = accepted_args

    async def execute(
        self,
        loop: asyncio.AbstractEventLoop,
        executor: ThreadPoolExecutor,
        evaluated_model_system_prompt: str | None,
        evaluated_model_retrieved_context: list[str] | None,
        evaluated_model_input: str | None,
        evaluated_model_output: str | None,
        evaluated_model_gold_answer: str | None,
        tags: dict[str, str] | None = None,
    ) -> TaskResult:
        kwargs = {
            "evaluated_model_system_prompt": evaluated_model_system_prompt,
            "evaluated_model_retrieved_context": evaluated_model_retrieved_context,
            "evaluated_model_input": evaluated_model_input,
            "evaluated_model_output": evaluated_model_output,
            "evaluated_model_gold_answer": evaluated_model_gold_answer,
            "tags": {**tags},
        }
        pass_kwargs = {k: v for k, v in kwargs.items() if k in self.accepted_args}
        result = await run_as_coro(
            __loop=loop,
            __executor=executor,
            __fn=self.task,
            **pass_kwargs,
        )
        if isinstance(result, str):
            return TaskResult(evaluated_model_output=result)
        return result

    @abc.abstractmethod
    def task(self, **kwargs) -> TaskResultT | str | typing.Awaitable[TaskResultT | str]: ...


class FunctionalTask(Task):
    fn: TaskFn

    def __init__(self, fn: TaskFn, accepted_args: set[str]):
        self.fn = fn
        super().__init__(accepted_args)

    def task(self, **kwargs) -> TaskResultT | str:
        return self.fn(**kwargs)


def task(fn: TaskFn) -> Task:
    sig = inspect.signature(fn)
    param_keys = sig.parameters.keys()
    for name in param_keys:
        if name not in TASK_ARGS:
            raise ValueError(f"{name!r} is not a valid task argument. Valid arguments are: {TASK_ARGS}")
    return FunctionalTask(fn, set(param_keys))


def simple_task(lambda_fn: typing.Callable[[str], str]) -> Task:
    @task
    def wrapper(evaluated_model_input: str) -> str:
        return lambda_fn(evaluated_model_input)

    return wrapper


@task
def nop_task(evaluated_model_output: str) -> str:
    return evaluated_model_output
